#######################################################################
###### R Class 'FJObj' was Designed and Developed by
###### BDBI, Product Innovation.
###### Authored by Joshua Luthy
###### Based on the insideFlow R package https://github.com/jluthy/insideFlow
###### This template was last updated in April 2024
#######################################################################
runID <- "FJ_RUNID"
#######################################################################
###### Create the FlowJo Object
#######################################################################
setClass("FJObj",
         slots = c(
           runID = "character",
           inputCSV = "list",
           parameterNames = "list",
           flowFrameFJ = "list",
           BatchCorrect = "list",
           PeacoQCPops = "list",
           FlowAIPops = "list",
           Rphenograph = "list",
           FastPhenograph = "list",
           FlowSOMClusters = "list",
           EmbedSOMReduction = "list",
           ClustRCheck = "list",
           Taylor = "list"
         ),
         prototype = list(
           runID = "runID",
           inputCSV = list(X = matrix(),
                           eventNumbers = matrix(),
                            cellIds = matrix()
           ),
           parameterNames = list(fjParNames = NULL,
                                 fjStainNames = NULL

           ),
           flowFrameFJ = list(ffFJ = list()
           ),
           BatchCorrect = list(uncorrected = matrix(),
                               corrected.asinh.transformed = matrix(),
                               corrected.reverse.transformed = matrix(),
                               markersToNorm = "character",
                               labels = matrix(),
                               id = matrix(),
                               colNames = matrix()

           ),
           PeacoQCPops = list(results = matrix()
           ),
           FlowAIPops = list(goodEvents = matrix(),
                             badEvents = matrix()
           ),
           Rphenograph = list(clusters = matrix(),
                              clustNumber = matrix(),
                              modularity = matrix()
           ),
           FastPhenograph = list(cluster = matrix(),
                                 clustNumber = matrix(),
                                 modularity = matrix()
           ),
           FlowSOMClusters = list(clusters = matrix(),
                                  clustNumber = matrix(),
                                  som = list()
           ),
           EmbedSOMReduction = list(embeddingMat = matrix(),
                               embeddingType = "character",
                               map = "ANY"
           ),
           ClustRCheck = list(values = matrix()
           ),
           Taylor = list(values = matrix()
           )
         ),
)
#######################################################################
##         Basic Slot Queries     ##
#######################################################################
setGeneric("getRunID", function(object) {
  standardGeneric("getRunID")
})

setMethod("getRunID",
          "FJObj",
          function(object){
            runid <- object@runID
            return(runid)
          }
)

setGeneric("getParNames", function(object) {
  standardGeneric("getParNames")
})

setMethod("getParNames",
          "FJObj",
          function(object){
            if(is.null(object@parameterNames$fjParNames)) {
              print("Parameter Names: None detected")
              return(NULL)
            } else {
              parNames <- as.character(object@parameterNames$fjParNames)
              return(parNames)
            }
          })

setGeneric("getStainNames", function(object) {
  standardGeneric("getStainNames")
})

setMethod("getStainNames",
          "FJObj",
          function(object){
            if(is.null(object@parameterNames$fjStainNames)) {
              print("Stain Names: None detected")
              return(NULL)
            } else {
              stNames <- as.character(object@parameterNames$fjStainNames)
              return(stNames)
            }
          })

setGeneric("getEmbeddingInfo", function(object) {
  standardGeneric("getEmbeddingInfo")
})
setMethod("getEmbeddingInfo",
          "FJObj",
          function(object){
            if(is.null(object@EmbedSOMReduction$embeddingType)) {
              print("Embedding Type: None detected")
              return(NULL)
            } else {
              embdType <- object@EmbedSOMReduction$embeddingType
              embdDims <- dim(object@EmbedSOMReduction$embeddingMat)[2]
              deets <- list(type=embdType, nDims=embdDims)
              return(deets)
            }
          }
)
#######################################################################
##         Load a file from single csv file     ##
#######################################################################
#' @return Returns the newly created object
#' @param object A FJObj object
#' @param csvPath The path to the exported csv file from FlowJo
#' @param eventNumber The column name for event numbers or cell ids
#' @importFrom data.table fread
#' @importFrom data.table :=
#' @examples \dontrun{
#' inputCSVpath <- "/Documents/ExtNode.csv"
#' myobject <- loadFlowJoCSV(object, inputCSVpath, eventNumber = "EventNumberDP")}
setGeneric("loadFlowJoCSV", function(object, csvPath, eventNumber = "EventNumberDP", cellIdColumn = "CellId") {
  standardGeneric("loadFlowJoCSV")
})

setMethod("loadFlowJoCSV",
          "FJObj",
          function(object, csvPath, eventNumber = "EventNumberDP", cellIdColumn = "CellId") {
            # Ensure data.table is available
            if (!requireNamespace("data.table", quietly = TRUE)) {
              stop("data.table package is not available")
            }

            # Using fread from data.table
            data <- data.table::fread(csvPath, check.names = FALSE)

            # Initialize empty matrices for eventNumbers and cellIds
            eventNumbers <- matrix(nrow = nrow(data), ncol = 0)
            cellIds <- matrix(nrow = nrow(data), ncol = 0)

            # Handle specified eventNumber column
            if (eventNumber %in% names(data)) {
              eventNumbers <- as.matrix(data[[eventNumber]])
              data[, (eventNumber) := NULL] # Remove specified eventNumber column after extracting it
            } else {
              warning(paste("Column", eventNumber, "not found in the provided CSV. eventNumbers will be left empty."))
            }

            # Handle specified CellId column
            if (cellIdColumn %in% names(data)) {
              cellIds <- as.matrix(data[[cellIdColumn]])
              data[, (cellIdColumn) := NULL] # Remove specified CellId column after extracting it
            } else {
              warning(paste("Column", cellIdColumn, "not found in the provided CSV. cellIds will be left empty."))
            }

            # Extracting parameter and stain names
            colNames <- names(data)
            parNames <- sapply(colNames, function(x) unlist(strsplit(x, " :: "))[1])
            stNames <- sapply(colNames, function(x) {
              parts <- unlist(strsplit(x, " :: "))
              if(length(parts) > 1) parts[2] else NA
            })

            # Trim whitespace from column names
            names(data) <- trimws(names(data))

            # Update object slots
            object@parameterNames <- list(fjParNames = parNames, fjStainNames = stNames)
            object@inputCSV <- list(X = as.data.frame(data))

            # Include eventNumbers and cellIds if they exist
            if (ncol(eventNumbers) > 0) {
              object@inputCSV$eventNumbers <- as.data.frame(eventNumbers)
              colnames(object@inputCSV$eventNumbers) <- "EventNumberDP"
            }
            if (ncol(cellIds) > 0) {
              object@inputCSV$cellIds <- as.data.frame(cellIds)
              colnames(object@inputCSV$cellIds) <- "CellId"
            }

            return(object)
          }
)




################################################################
#    Prepare for batch corrections  #
################################################################
#' @title Transform data before batch correction step
#' @return Returns the object with asinh-transformed values to batch correct
#' @param object A FJObj object
#' @param fjMarkersList The selected list of markers to normalize
#' @param fjCofactor The selected Cofactor value to apply
#' @importFrom cyCombine transform_asinh
#' @examples \dontrun{
#' myobject <- prepareBEC(object, markers, cofact)}
setGeneric("prepareBEC", function(object, fjMarkersList, fjCofactor){
  standardGeneric("prepareBEC")
})
# Takes the object and list of markers and transforms the matrix in preperation
# for clustering and batch correction
setMethod("prepareBEC",
          "FJObj",
          function(object, fjMarkersList, fjCofactor){
            object@BatchCorrect$markersToNorm <- fjMarkersList

            object@BatchCorrect$uncorrected <- transform_asinh(
              df = as.data.frame(object@inputCSV$X),
              markers = object@BatchCorrect$markersToNorm,
              cofactor = fjCofactor,
              .keep = TRUE # Lets you keep all columns, in case they are useful to you
            )
            return(object)
          }
)
################################################################
#    Prepare for batch effect checks  #
################################################################
#' @title Transform data and clean up names before checking for batch effects
#' @return Returns the object with asinh-transformed values to batch correct
#' @param object A FJObj object
#' @param fjMarkersList The selected list of markers to normalize
#' @param fjCofactor The selected Cofactor value to apply
#' @param fjBatchID The selected batch keyword value to use
#' @param fjSamID The selected sample keyword value to use
#' @importFrom cyCombine transform_asinh
#' @examples \dontrun{
#' myobject <- prepareBECheck(object, markers, cofact, batchID)}
setGeneric("prepareBECheck", function(object, fjMarkersList, fjCofactor, fjBatchID, fjSamID){
  standardGeneric("prepareBECheck")
})
# Takes the object and list of markers and transforms the matrix in preperation for batch effect evaluations.
# Cleans up the data matrix by removing columns that are not in the selected markers list and uses Stain Names as column names.
setMethod("prepareBECheck",
          "FJObj",
          function(object, fjMarkersList, fjCofactor, fjBatchID, fjSamID){
            object@BatchCorrect$markersToNorm <- fjMarkersList

            object@BatchCorrect$uncorrected <- transform_asinh(
              df = as.data.frame(object@inputCSV$X),
              markers = object@BatchCorrect$markersToNorm,
              cofactor = fjCofactor,
              .keep = FALSE # Lets you keep all columns, in case they are useful to you
            )
            # Extract parameterNames and stainNames from the selected_markers list
            selected_parameters <- sapply(fjMarkersList, function(x) unlist(strsplit(x, " :: "))[1])
            selected_stains <- sapply(fjMarkersList, function(x) {
              parts <- unlist(strsplit(x, " :: "))
              if(length(parts) > 1) gsub("[^A-Za-z0-9]", ".", parts[2]) else NA
            })
            # Filter the columns based on selected markers
            selected_columns <- unlist(lapply(fjMarkersList, function(marker) {
              idx <- which(selected_parameters == unlist(strsplit(marker, " :: "))[1] &
                             selected_stains == gsub("[^A-Za-z0-9]", ".", unlist(strsplit(marker, " :: "))[2]))
              if (length(idx) == 0) return(NULL)
              idx
            }))
            # Remove NA columns (where stain name is not available)
            selected_columns <- selected_columns[!is.na(selected_columns)]
            # Subset the data matrix with selected columns
            object@BatchCorrect$uncorrected <- object@BatchCorrect$uncorrected[, selected_columns, drop = FALSE]
            # Set column names with stain names
            colnames(object@BatchCorrect$uncorrected) <- selected_stains
            # Set the batch column
            object@BatchCorrect$uncorrected$batch <- as.factor(object@inputCSV$X[,fjBatchID])
            # Set the sample column
            object@BatchCorrect$uncorrected$sample <- as.factor(object@inputCSV$X[,fjSamID])

            return(object)
          }
)
################################################################
#    Perform Batch Correction  #
################################################################
#' @title Performs batch correction with cyCombine
#' @return Returns the batch corrected object
#' @param object A FJObj object
#' @param batchID The selected list of markers to normalize
#' @param normMethod The selected Cofactor value to apply
#' @param xdim The first dimension of SOM used
#' @param ydim The second dimension of SOM used
#' @param covariate The optional covariate to use in the batch correction
#' @importFrom cyCombine normalize
#' @importFrom cyCombine create_som
#' @importFrom magrittr %>%
#' @examples \dontrun{
#' myobject <- runBatch_correct(object, batchID, nrmMethod, xdim, ydim)}
#' @docType methods
#' @export
setGeneric("runBatch_correct", function(object, batchID, normMethod, xdim, ydim, covariate){
  standardGeneric("runBatch_correct")
})

setMethod("runBatch_correct",
          "FJObj",
          function(object, batchID, normMethod, xdim, ydim, covariate){
            covariate <- evaluate_fjCovariate(covariate)  # Process covariate parameter
            object@BatchCorrect$uncorrected$batch <- as.factor(object@BatchCorrect$uncorrected[,batchID])
            # Run batch correction
            object@BatchCorrect$labels <- object@BatchCorrect$uncorrected %>%
              cyCombine::normalize(markers = object@BatchCorrect$markersToNorm,
                                   norm_method = normMethod) %>%
              cyCombine::create_som(markers = object@BatchCorrect$markersToNorm,
                                    rlen = 10,
                                    xdim = xdim,
                                    ydim = ydim)

            # 'covar' is the optional second keyword present in the exported data.
            if (!is.null(covariate)) {
              object@BatchCorrect$uncorrected[,covariate] <- as.factor(object@BatchCorrect$uncorrected[,covariate])
              object@BatchCorrect$corrected.asinh.transformed <- object@BatchCorrect$uncorrected %>%
                correct_data(label = object@BatchCorrect$labels,
                             markers = object@BatchCorrect$markersToNorm,
                             covar = covariate)
            } else {
              object@BatchCorrect$corrected.asinh.transformed <- object@BatchCorrect$uncorrected %>%
                correct_data(label = object@BatchCorrect$labels,
                             markers = object@BatchCorrect$markersToNorm)
            }
            return(object)
          }
)
################################################################
#    Prepare for CyCombine Image Creation  #
################################################################
#' @title Return plots comparing markers after batch normalization
#' @return Returns the plots from cyCombine showing normalized histograms
#' @param object A FJObj object
#' @param fjPopName The name of selected population in FlowJo
#' @param fjCofactor The selected Cofactor value to apply
#' @param outPutFolder The path to the output folder
#' @importFrom cyCombine plot_density
#' @examples \dontrun{
#' normalizedPlots(object, popName, coFact, outsFolder)}
setGeneric("normalizedPlots", function(object, fjPopName, fjCofactor, outPutFolder){
  standardGeneric("normalizedPlots")
})

setMethod("normalizedPlots",
          "FJObj",
          function(object, fjPopName, fjCofactor, outPutFolder){
            if(is.null(myFJobj@BatchCorrect$uncorrected$label)){
              print("Adding cluster labels")
              myFJobj@BatchCorrect$uncorrected$label <- myFJobj@BatchCorrect$corrected.asinh.transformed$label
            }else{print("Cluster labels exist")}
            if(is.null(myFJobj@BatchCorrect$uncorrected$id)){
              print("Adding cell ids")
              myFJobj@BatchCorrect$uncorrected$id <- myFJobj@BatchCorrect$corrected.asinh.transformed$id
            }else{print("Cell ids exist")}

            # Plots can be made like so:
            # png("NUL")
            # pdf("NUL")
            plot_density(uncorrected = myFJobj@BatchCorrect$uncorrected,
                         corrected = myFJobj@BatchCorrect$corrected.asinh.transformed,
                         markers = myFJobj@BatchCorrect$markersToNorm,
                         filename = paste0(outPutFolder, "Density_plot_", fjPopName, "_cofactor_" , fjCofactor,"_.png" ))

            return(object)
          })
################################################################
#    Reverse Transform before export to fcs  #
################################################################
#' @title Reverse transforms data matrix before exporting to FlowJo
#' @return Returns reversed transformed data matrix before export to FlowJo
#' @param object A FJObj object
#' @param fjCofactor The selected Cofactor value to apply
#' @importFrom cyCombine transform_asinh
#' @examples \dontrun{
#' myobject <- reverseTransform(object, popName, coFact, outsFolder)}
setGeneric("reverseTransform", function(object, fjCofactor){
  standardGeneric("reverseTransform")
})

setMethod("reverseTransform",
          "FJObj",
          function(object, fjCofactor){
            # Transform back to scaled values before export
            object@inputCSV$X <- transform_asinh(
              df = object@BatchCorrect$corrected.asinh.transformed,
              markers = object@BatchCorrect$markersToNorm,
              cofactor = fjCofactor,
              reverse = TRUE,
              .keep = TRUE # Lets you keep all columns, in case they are useful to you
            )
            for (col in names(object@inputCSV$X)) {
              # Check if the column is a factor
              if (is.factor(object@inputCSV$X[[col]])) {
                # Convert factor to integer
                object@inputCSV$X[[col]] <- as.integer(object@inputCSV$X[[col]])
              }
            }
            return(object)
          })

#######################################################################
##         Cleanup Object before exporting FCS post cyCombine        ##
#######################################################################
#' @title Cleanup Object
#' @return Returns the light-weight object cleaned of transformed data slots
#' from batch correction processes
#' @param object A FJObj object
#' @examples \dontrun{
#' myobject <- cleanupFJobj(object)}
setGeneric("cleanupFJObj", function(object){
  standardGeneric("cleanupFJObj")
})
setMethod("cleanupFJObj",
          "FJObj",
          function(object){
            if(!is.null(object@inputCSV$X)){
              # remove added cols from cyCombine
              if("id" %in% colnames(object@inputCSV$X)){
                object@inputCSV$X[["id"]] <- NULL
              }
              if("label" %in% colnames(object@inputCSV$X)){
                object@inputCSV$X[["label"]] <- NULL
              }
              if("batch" %in% colnames(object@inputCSV$X)){
                object@inputCSV$X[["batch"]] <- NULL
              }
              # Split the column names at "::" and take the first part
              split_colnames <- sapply(strsplit(colnames(object@inputCSV$X), " :: "), `[`, 1)
              # get the original ordering before exporting
              paramOrderIdx <- match(object@parameterNames$fjParNames, split_colnames)
              # Replace original expression matrix with batch corrected one
              object@inputCSV$X <- object@inputCSV$X[ ,paramOrderIdx]

              # Clean up to make the object light
              object@BatchCorrect$corrected.asinh.transformed <- as.matrix(0)
              object@BatchCorrect$corrected.reverse.transformed <- as.matrix(0)
              print("cleaned FlowJo R object")
            }
            return(object)
          }
)
################################################################
#    Build a FlowFrame  #
################################################################
## Generates GUID like "47689979-e471-47a6-aa1a-7910d2d42b03"
## As a side effect, this also sets the seed based in the data passed in
generateGUID <- function(data) {
  mySeed <- round(1000*mean(as.matrix(data))) + dim(data)[1] + dim(data)[2]
  while (mySeed > .Machine$integer.max) mySeed <- round(mySeed/3)
  set.seed(as.integer(mySeed))

  part1 <- paste(unlist(lapply(1:8, function(x) { sprintf("%x", round(runif(1, 0, 9))) })), collapse="", sep="")
  part2 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part3 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part4 <- paste(unlist(lapply(1:4, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  part5 <- paste(unlist(lapply(1:12, function(x) { sprintf("%x", round(runif(1, 0, 15))) })), collapse="", sep="")
  paste(part1, part2, part3, part4, part5, sep="-")
}

#' @title Build a Flowframe from CSV file expression matrix
#' @return Returns reversed transformed data matrix before export to FlowJo
#' @param object A FJObj object
#' @param dropCompPrefix Boolean to allow for paramName cleanup
#' @importFrom methods new
#' @examples \dontrun{
#' myobject <- flowFrameFromFlowJoCSV(object, popName, coFact, outsFolder)}
#' @docType methods
#' @export
setGeneric("flowFrameFromFlowJoCSV", function(object, filename, dropCompPrefix = FALSE) {
  standardGeneric("flowFrameFromFlowJoCSV")
})

setMethod("flowFrameFromFlowJoCSV",
          "FJObj",
          function(object, filename, dropCompPrefix=FALSE){
            data <- as.data.frame(object@inputCSV$X)
            colNames <- colnames(data)
            npar <- length(colNames)
            id <- paste("$P",1:npar,sep="")

            maxRange <- unlist(lapply(1:npar, function(x) {
              max(data[,x])
            }))
            range <- maxRange + 1
            minRange <- unlist(lapply(1:npar, function(x) {
              min(0, min(data[,x]))
            }))

            colNames <- gsub("^\\s+|\\s+$", "", colNames) 	# Trim the leading and tailing whitespaces

            ## FlowJo splits parameter names and description by " :: ", e.g., "Comp-PE-Cy5-A :: CD19#PE-Cy5"
            names <- unlist(lapply(1:npar, function(i) {
              unlist(strsplit(colNames[i], " :: "))[1]
            }))
            ## FlowJo indicates that a patameter is compensated by the Comp- prefix in its name, we can drop that if undesirable
            if (dropCompPrefix) {
              names <- gsub("^\\Comp-", "FJNorm-", names)
            }
            ## Again splitting by " :: " but this time taking the second part (if present) as description
            desc <- unlist(lapply(1:npar, function(i) {
              s <- unlist(strsplit(colNames[i], " :: "))
              if (length(s) >= 2) s[2]
              else NA
            }))

            pars <- new("AnnotatedDataFrame",
                        data=data.frame(
                          row.names=I(id), name=I(names), desc=I(desc),
                          range=range, minRange=minRange, maxRange=maxRange),
                        varMetadata=data.frame(
                          row.names=I(c("name","desc","range", "minRange", "maxRange")),
                          labelDescription=I(c("Name of Parameter", "Description of Parameter", "Range of Parameter", "Minimum Parameter Value after Transforamtion", "Maximum Parameter Value after Transformation"))
                        )
            )

            ## Build the descripton
            txt <- list()
            txt["FCSversion"] <- "3"
            txt["$PAR"] <- as.character(npar)
            txt["$TOT"] <- as.character(dim(data)[1])
            txt["$BYTEORD"] <- "4,3,2,1"
            txt["$DATATYPE"] <- "F"
            txt["$MODE"] <- "L"
            txt["$NEXTDATA"] <- "0"

            txt["$BEGINANALYSIS"] <- "0"
            txt["$ENDANALYSIS"] <- "0"
            txt["$BEGINSTEXT"] <- "0"
            txt["$ENDSTEXT"] <- "0"
            dataStart <- 1000 ## This doesn't really matter, it's not an FCS file anyway, but let's provide something reasonable
            txt["$BEGINDATA"] <- as.character(dataStart)
            txt["$ENDDATA"] <- as.character(dataStart + npar * dim(data)[1] * 4 - 1) ## $DATATYPE is F and PnB is 32, so here we do: start + rows x cols x 4 bytes per value - 1

            txt[paste(id,"N",sep="")] <- names
            txt[paste(id,"S",sep="")] <- desc
            txt[paste(id,"R",sep="")] <- as.character(range)
            txt[paste(id,"B",sep="")] <- "32"
            txt[paste(id,"E",sep="")] <- "0,0"
            txt[paste(id,"G",sep="")] <- "1"
            txt[paste("flowCore_",id,"Rmin",sep="")] <- as.character(minRange)
            txt[paste("flowCore_",id,"Rmax",sep="")] <- as.character(maxRange)

            txt["$FIL"] <- basename(filename)
            txt["FILENAME"] <- object@runID
            txt["$SYS"] <- paste(Sys.info()['sysname'], Sys.info()['release'])
            txt["CREATOR"] <- R.version.string
            txt["transformation"] <- "applied"
            txt["GUID"] <- generateGUID(data)
            txt["ORIGINALGUID"] <- generateGUID(data)
            txt["FJRUNID"] <- object@runID

            txt["$INST"] <- "Unknown institution"
            txt["$OP"] <- "Unknown operator"
            txt["$CYT"] <- "Unknown cytometer"
            txt["$SRC"] <- "Unknown source"

            curtime <- Sys.time()
            txt["$DATE"] <- toupper(format(curtime, "%d-%b-%Y")) ## Current data in the format of "05-OCT-2017"
            timeDelta <- 120 ## picking 2 minutes as cytometer run time
            if ("time" %in% tolower(names)) {
              txt["$TIMESTEP"] <- "0.01" ## We don't know that, so just something that would be reasonable-ish
              timeIndex <- which(tolower(names) == "time")
              timeDelta <- ceiling((max(data[,timeIndex]) - min(data[,timeIndex])) * 0.01)
            }
            if (toupper(format(curtime - timeDelta, "%d-%b-%Y")) == toupper(format(curtime, "%d-%b-%Y"))) {
              ## Subtracting timeDelta is still the same day, so let's do it that way.
              txt["$BTIM"] <- format(curtime - timeDelta, "%H:%M:%S")
              txt["$ETIM"] <- format(curtime, "%H:%M:%S")
            } else {
              ## Subtracting timeDelta gets us to yesterday, that's not good, so let's just say we started at midnight and add delta to it as ETIM...
              ## (all this is arbitrary anyway, we just want resonable values)
              txt["$BTIM"] <- "00:00:00"
              txt["$ETIM"] <- sprintf("%02i:%02i:%02i", floor(timeDelta / 3600), floor((timeDelta%%3600) / 60), floor(timeDelta%%60))
            }
            colnames(data) <- names
            object@flowFrameFJ <- list(fcs=new("flowFrame",
                                               exprs=as.matrix(data),
                                               parameters=pars,
                                               description=txt),
                                       cellIds=object@inputCSV$eventNumbers)

            return(object)
          })


################################################################
#    Perform Batch Correction  #
################################################################
#' Full function for detection of batch effects using cluster proportions
#'
#' This function is used for batch effect detection in multidimensional datasets.
#'   The function applies a SOM-based clustering to a dataset in order to compare not only marker expression differences across batches,
#'   but also the cluster percentages in each batch to detect possible populations that are over-/under-represented in a single batch.
#'   This is coupled with UMAP plots to assist the interpretation of the results.
#'   However, this is primarily meaningful for sets with 3-30 batches - in cases outside this range, only the UMAPs will be generated.
#'
#' @param df Tibble containing the expression data and batch information. See prepare_data.
#' @param downsample Number of cells to include in detection. If not specified all cells will be used. One should be careful with the downsampling here as too strong downsampling leads to spurious results.
#' @param norm_method Normalization methods (options = 'scale' and 'rank')
#' @param xdim Grid size in x-axis for SOM (default = 8)
#' @param ydim Grid size in y-axis for SOM (default = 8)
#' @param seed Random seed for reproducibility
#' @param markers If only some markers should be used this parameter is used to define them. If not set, all markers are used.
#' @param batch_col Name of column containing batch information
#' @param label_col If existing labels should be used, this column must be present in the data
#' @param out_dir Directory for plot output
#' @param name Name of dataset - used for plot titles
#' @examples
#' \dontrun{
#' detect_batch_effect(df = FJObj@BatchCorrect$uncorrected)
#' detect_batch_effect(df = FJObj@BatchCorrect$uncorrected, xdim = 8, ydim = 8, seed = 23,
#'                     markers = c('CD3', 'CD4', 'CD8a', 'CD20', 'CD19', 'CD56', 'CD33'))
#' }
setGeneric("detect_batch_effect_JL",
           function(object,
                    out_dir,
                    downsample = NULL,
                    norm_method = "scale",
                    xdim = 8,
                    ydim = 8,
                    seed = 23,
                    markers = NULL,
                    batch_col = "batch",
                    label_col = "label",
                    name = 'raw data',
                    plotType = "umap") {
             standardGeneric("detect_batch_effect_JL")
           })

setMethod("detect_batch_effect_JL",
          "FJObj",
          function(object,
                   out_dir,
                   downsample = NULL,
                   norm_method = "scale",
                   xdim = 8,
                   ydim = 8,
                   seed = 23,
                   markers = NULL,
                   batch_col = "batch",
                   label_col = "label",
                   name = 'raw data',
                   plotType = "umap") {

            df <- object@BatchCorrect$uncorrected
            # Create output directory if missing
            cyCombine:::check_make_dir(out_dir)

            # Initialize an empty list to store the messages
            messages <- list()

            # Define a function to both print and store a message
            store_message <- function(msg) {
              message(msg)
              messages <<- c(messages, list(msg))
            }

            # Check batch_col and rename if necessary
            cyCombine:::check_colname(colnames(df), batch_col, location = "df")
            if (batch_col != 'batch') {
              df$batch <- df[, batch_col]
            }
            df$batch <- as.factor(df$batch)

            # Check multiple batches
            if (length(levels(df$batch)) < 2) {
              stop('Error! Please provide a datasets with more than one batch.')
            }

            # Check out dir
            if (is.null(out_dir)) {
              stop('Error! Please speicify output directory.')
            }

            # Downsample if requested
            if (!is.null(downsample)) {
              message(paste0('Downsampling to ', downsample, ' cells.'))
              if (downsample <= nrow(df)) {
                set.seed(seed)
                df <- df %>% dplyr::slice_sample(n = downsample)
              } else {
                warning('Specified downsample parameter exceeds the number of cells in the dataset.')
              }
            }

            if (is.null(markers)){
              markers <- df %>%
                cyCombine::get_markers()
            }

            # Create SOM on scaled data if no labels exist
            if (!(label_col %in% colnames(df))) {
              message('Determining new cell type labels using SOM:')
              labels <- df %>%
                cyCombine::normalize(markers = markers,
                                     norm_method = norm_method) %>%
                cyCombine::create_som(markers = markers,
                                      seed = seed,
                                      xdim = xdim,
                                      ydim = ydim)

              # Get SOM output
              df$label <- labels

            } else {
              message('Using existig cell type labels.')
              df$label <- df[[label_col]]

            }

            # Perform EMD and cluster percentage calculations - and filter to get outliers (only for sets with 3-30 batches)
            if (length(levels(df$batch)) >= 3 & length(levels(df$batch)) <= 30) {
              emd <- df %>%
                dplyr::mutate(label = as.character(label)) %>%
                cyCombine::compute_emd()

              markers_emd <- list()
              # Looking through EMDs to find culprits - using loops
              for (m in markers) {

                # Calculation of the mean per batch across all clusters for each marker
                marker_emd <- lapply(emd, function(x) {x[[m]]})
                marker_emd <- lapply(marker_emd, function(x) {Matrix::forceSymmetric(x, uplo='U')})

                marker_emd <- do.call(rbind, marker_emd)
                marker_emd <- Matrix::colMeans(as.matrix(marker_emd), na.rm = T)

                markers_emd[[m]] <- marker_emd
              }

              marker_emd <- which(stats::p.adjust(sapply(markers_emd, function(x) {outliers::dixon.test(x, opposite=F)$p.value}), method = 'BH') < 0.05 | stats::p.adjust(sapply(markers_emd,  function(x) {outliers::dixon.test(x, opposite=T)$p.value}), method = 'BH') < 0.05)
              store_message(paste0('\nThere are ', length(marker_emd), ' markers that appear to be outliers in a single batch:'))
              store_message(paste(markers[marker_emd], collapse = ', '))

              # Look for cluster over- and under-representation
              counts <- table(df$batch, df$label)
              perc <- (counts / Matrix::rowSums(counts))*100

              # use the Dixon test to look for outliers (test whether a single low or high value is an outlier)
              out_cl <- which(stats::p.adjust(apply(perc, 2, function(x) {outliers::dixon.test(x, opposite=F)$p.value}), method = 'BH') < 0.05 | stats::p.adjust(apply(perc, 2, function(x) {outliers::dixon.test(x, opposite=T)$p.value}), method = 'BH') < 0.05)

              store_message(paste0('\nThere are ', length(out_cl), ' clusters, in which a single cluster is strongly over- or underrepresented.'))

              for (cl in out_cl) {
                store_message(paste0('The cluster percentages for each batch in cluster ', cl, ' are:'))
                store_message(paste(names(perc[,cl]), '=', round(perc[,cl], 2), '%', collapse = ', '))

                exp_markers <- df %>%
                  dplyr::filter(label == cl) %>%
                  dplyr::summarise(dplyr::across(cyCombine::get_markers(df), stats::median)) %>%
                  unlist() %>%
                  sort(decreasing = T)

                store_message(paste0('The cluster expresses ', paste(names(exp_markers[which(exp_markers > 1)]), collapse = ', ')))
                cat('\n')
              }
            } else {
              store_message('The number of batches is not within the range of 3-30, so only UMAP plots will be generated.')
            }
            # Making a set of UMAP plots to help the diagnostics
            message('Making UMAP plots for up to 50,000 cells.')

            # SHOULD WE DOWNSAMPLE??
            if (nrow(df) > 50000) {
              set.seed(seed)
              df <- df %>% dplyr::slice_sample(n = 50000)
            }

            plot_dimred_full_JL(df,
                                name,
                                type = plotType,
                                markers = NULL,
                                seed = 23,
                                out_dir)

            message(paste0('Saved UMAP plot for batches and labels here: ', out_dir, ' as UMAP_batches_labels.png.'))
            message(paste0('Saved UMAP plot colored by each marker in directory: ', out_dir, 'UMAP_markers.\n'))
            # Write the messages to a text file
            messages <- lapply(messages, as.character)
            writeLines(unlist(messages), con = file.path(out_dir, "Batch_Effect_Evaluation.txt"))
            message('Done!')
          })


#######################################################################
#   Generate Batch Effect Evaluation Plots with fast EmbedSOM plots   #
#######################################################################
#' Dimensionality reduction plots - colored with labels, batches and marker expression
#' @inheritParams plot_dimred
#' @param out_dir Directory to put output figures
#' @param name Name of the dataset
plot_dimred_full_JL <- function(df,
                                name,
                                type = "umap",
                                markers = NULL,
                                seed = 23,
                                out_dir = NULL) {

  # Check out dir
  if (is.null(out_dir)) {
    stop('Error! Please speicify output directory.')
  } else{
    # Create output directory if missing
    cyCombine:::check_make_dir(out_dir)
  }

  if(is.null(markers)){
    markers <- cyCombine::get_markers(df)
  }
  Batch <- df$batch %>%
    as.factor()

  Label <- df$label %>%
    as.factor()

  df <- df %>%
    dplyr::select(dplyr::all_of(markers))

  if (type == "tsne") {
    # Run tSNE
    set.seed(seed)
    map <- EmbedSOM::RandomMap(df, min(5000, nrow(df)), coordsFn=EmbedSOM::tSNECoords())
    embed <- EmbedSOM::EmbedSOM(df, map=map, parallel=T, k=25, adjust=0, smooth=0)

    # Make dataframe with output
    df <- cbind.data.frame(embed, as.factor(Batch), as.factor(Label), df); colnames(df) <- c("ES.TSNE1", "ES.TSNE2", "Batch", "Label", markers)

  } else if (type == "umap") {
    # Run UMAP
    set.seed(seed)
    map <- EmbedSOM::RandomMap(df, min(20000, nrow(df)), coordsFn=EmbedSOM::uwotCoords(min_dist=0.5))
    embed <- EmbedSOM::EmbedSOM(df, map=map, parallel=T, k=25, adjust=0, smooth=0)

    # Make dataframe with output
    df <- cbind.data.frame(embed, as.factor(Batch), as.factor(Label), df); colnames(df) <- c("ES.UMAP1", "ES.UMAP2", "Batch", "Label", markers)
  }

  # Make the plots
  batch_plot <- df %>%
    ggplot2::ggplot(ggplot2::aes_string(x = colnames(df)[1],
                                        y = colnames(df)[2])) +
    ggplot2::geom_point(ggplot2::aes_string(color = "Batch"),
                        alpha = 0.3,
                        size = 0.4,
                        shape = 1) +
    ggplot2::guides(color = ggplot2::guide_legend(override.aes = list(alpha = 1, size = 1))) +
    ggplot2::theme_bw() + ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5)) +
    ggplot2::ggtitle(paste(toupper(type), "-", name))

  label_plot <- df %>%
    ggplot2::ggplot(ggplot2::aes_string(x = colnames(df)[1],
                                        y = colnames(df)[2])) +
    ggplot2::geom_point(ggplot2::aes_string(color = "Label"),
                        alpha = 0.3,
                        size = 0.4,
                        shape = 1) +
    ggplot2::guides(color = ggplot2::guide_legend(override.aes = list(alpha = 1, size = 1))) +
    ggplot2::theme_bw() + ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5)) +
    ggplot2::ggtitle(paste(toupper(type), "-", name))

  # Saving plots
  if (type == "tsne") {
    cyCombine::plot_save_two(batch_plot, label_plot, paste0(out_dir, '/tSNE_batches_labels.png'))
  } else if (type == "umap") {
    cyCombine::plot_save_two(batch_plot, label_plot, paste0(out_dir, '/UMAP_batches_labels.png'))
  }


  # Marker plots
  if (type == "tsne") {
    cyCombine:::check_make_dir(paste0(out_dir, '/tSNE_markers'))
  } else if (type == "umap"){
    cyCombine:::check_make_dir(paste0(out_dir, '/UMAP_markers'))
  }

  marker_plots <- list()
  for (m in markers) {
    p <- ggplot2::ggplot(df, ggplot2::aes_string(x = colnames(df)[1],
                                                 y = colnames(df)[2])) +
      ggplot2::geom_point(ggplot2::aes_string(color = m),
                          alpha = 0.3,
                          size = 0.4) +
      ggplot2::scale_color_gradientn(m, colors = grDevices::colorRampPalette(rev(RColorBrewer::brewer.pal(n = 11, name = "Spectral")))(50)) +
      ggplot2::theme_bw() + ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0.5)) +
      ggplot2::ggtitle(paste(toupper(type), "-", name))
    if (type == "tsne") {
      suppressMessages(ggplot2::ggsave(p, filename = paste0(out_dir, '/tSNE_markers/tSNE_batches_labels', m, '.png')))
    } else if (type == "umap") {
      suppressMessages(ggplot2::ggsave(p, filename = paste0(out_dir, '/UMAP_markers/UMAP_batches_labels', m, '.png')))
    }

  }

}

evaluate_fjCovariate <- function(fjCovariate) {
  if (fjCovariate == "None") {
    fjCovariate <- NULL
  }
  return(fjCovariate)
}
